(* Time-stamp: <modified the 03/05/2018 (at 11:06) by Erwan Jahier> *)
(* An experimental wrapper around rdbg-top to 
    - avoid typing "()" and ";;" 
    - generate initial sessions
    - manage/load previous sessions
*)
open RdbgWrapperArgs

let add_semi_ref = ref true

let rdbg_top = "rdbg-top" 


let url_doc="http://www-verimag.imag.fr/DIST-TOOLS/SYNCHRONE/lustre-v6/"
              
let _ =
  lurette := false;
  usage := (Sys.argv.(0) ^
               " is an experimental wrapper around the ocaml toplevel rdbg-top.
rdbg basically performs 3 things.

(1) It generates a rdbg-session-<i>.ml file template (if required)
    that is loaded before starting.

For example, try:
  rdbg -sut \"lv6 file.lus -n node\" -env \"lutin f.lut -n main\"
  rdbg -sut \"ecexe-rif f.ec node\" -env \"lutin f.lut -n main\"
  lv6 -2c -cc f.lus -n n ;rdbg -sut \"./n.exec\" -env \"lutin f.lut -n main\"

(2) It  generates a default  '"^rdbg_tuning_file^"' file used  to contain
    user commands,  and that  is loaded  before starting.   rdbg won't
    override '"^rdbg_tuning_file^"'  if it  already exists.   The --emacs
    option generates a version that contain emacs support (experimental).

(3) It  discharges users from  writting \"()\" and \";;\"  before rdbg
    commands.  This wrapping is
     - turned on/off by the '#auto_on'/'#auto_off directives';
     - locally inhibited for lines beginning with the ' ' character.

    Hint: preprend rdbg by 'ledit' (or 'ledit -h rdbg-history -x').

cf also "^url_doc^"

usage: [ledit] "^ Sys.argv.(0) ^" [<options>] [<ocaml files>]
where <ocaml files> (e.g., 'rdbg-session-i.ml') are evaluated by rdbg-top
      <options> are:")


let get_extension f =
  try
    let bf = Filename.chop_extension f in
    let lf,lbf = String.length f, String.length bf in
    String.sub f (lbf+1) (lf - lbf - 1)
  with _ -> ""

let args= ref []
let ml_files = ref []

let add_args arg = 
  if get_extension arg = "ml" 
  then 
    ml_files := arg::!ml_files 
  else
    args := (arg::!args)

let others = ref []
let (add_other : string -> unit) =
  fun s -> 
    others := s:: !others

let other_args:string =
  mkoptab ();
  Arg.parse_argv Sys.argv !arg_options add_args !usage;
  String.concat " " (List.rev !others)

let rdbg_top_cmd = rdbg_top ^ " -I +rdbg-plugin " ^ other_args
  
(**********************************************************************************)
(* generete a rdbg-session.ml *)

let prolog =  ((Mypervasives.entete2 "(*" "*)" RdbgVersion.str RdbgVersion.sha) ^
                 (if !lurette then "" else "
#use \"topfind\";; 
#require \"rdbg-plugin\";;")^("
open Event;;
open RdbgArg;;
open RdbgRun;;
open RdbgMain;;
open RdbgStdLib;;
open Data;;
"))

let (tool_to_ocaml_pack_name: string -> string) = function
  | "lutin"   -> "lutin"
  | "lv6" -> "lustre-v6"
  | "lus2lic" -> "lustre-v6"
  | _ -> "stdio"

let gen_plugin_def packs_ref oc nodbg name (str:string) =
  let args = Str.split (Str.regexp "[ \t]+") str in
  let tool = Filename.basename (List.hd args) in
  let make_str,opt_forced =
    match tool with
    | "lutin"   -> "LutinRun.make",[]
    | "lus2lic" -> "Lv6Run.make",["--expand-io-type"] 
    | "lv6" ->
       (* necessary because lv6 -2c generates rif where all i/o are expanded.*)
        "Lv6Run.make",["--expand-io-type"] 
    | _ -> "StdioRun.make",[]
  in
  let args = args @ opt_forced in
  let args_str = String.concat ";" (List.map (fun s -> "\""^s^"\"") args) in
  let pn = tool_to_ocaml_pack_name tool in
  if not (List.mem pn !packs_ref || pn = "stdio") then packs_ref := pn::!packs_ref ;
  if (not !lurette && pn <> "stdio") then Printf.fprintf oc "#require \"%s\";;\n" pn;
  if pn = "stdio" then (
    Printf.fprintf oc "\nlet %s = 
  let plugin = StdioRun.make \"%s\" in
  plugin\n;;\n" name (String.concat " " args);
    flush oc
  ) else (
    Printf.fprintf oc "\nlet %s = 
  let args = [%s] in
  let aargs = Array.of_list args in
  let plugin = %s aargs in
  %s\n;;\n" name args_str make_str
                   (if nodbg then "let skip_dbg sl e cont = cont (plugin.step sl) e in
  { plugin with step_dbg = skip_dbg }"
                    else "plugin")
                   
  ;
    flush oc)

let gen_arg_fill oc rp plugin_list =
    Printf.fprintf 
      oc "let _ = args.%s <- [%s];;\n" 
      rp
      (String.concat ";" (List.map (fun n-> "Ocaml("^n^")") plugin_list))
 
let rdbg_tuning ()=
  (if !emacs_mode then 
  "let hl hook (file, line, charb, chare) =
  let cmd = Printf.sprintf \"emacsclient -e '(save-selected-window (find-file-other-window \\\"%s\\\") %s (goto-char (point-min)) (forward-line %i)  (forward-char %i) (set-mark-command nil) (forward-char %i) (hlt-highlight-region) )' \"
 file hook (line-1) charb (max 1 (chare-charb))
  in
  ignore(Sys.command cmd)

let hl_atoms = function
  | []-> ()
  | x::t -> hl \"(hlt-unhighlight-region)\" x; List.iter (hl \"\") t
        
let _ = 
  let cmd = \"emacsclient -e ' (hlt-choose-default-face \\\"org-checkbox\\\") '\" in
  ignore(Sys.command cmd)

let emacs_udate e =
  try 
  match e.sinfo with
  | None -> ()
  | Some si ->
     let si_list =
       List.map
         (fun si -> (si.file, fst si.line, fst si.char, snd si.char))
         (si()).atoms
     in
     hl_atoms si_list
  with
    Match_failure _  -> ()
      
let _ = time_travel true;; 
let e = ref (RdbgStdLib.run());;
let s () = e:=step !e; ignore(emacs_udate !e);;
let si i = e:=stepi !e i; ignore(emacs_udate !e);;
let n () = e:=next !e; ignore(emacs_udate !e);;
let ni i = e:= nexti !e i;emacs_udate !e;;
let g i = e:= goto !e i;emacs_udate !e;;
let b () = e:= back !e ;emacs_udate !e;;
let bi i = e:= backi !e i;emacs_udate !e;;
let r () = e:=RdbgStdLib.run();emacs_udate !e;;
let c () = e:= continue !e;emacs_udate !e;;
"
  else 
      "
let _ = time_travel true;; 
let e = ref (RdbgStdLib.run());;
let s () = e:=step !e;;
let si i = e:=stepi !e i;;
let n () = e:=next !e  ;;
let ni i = e:= nexti !e i;;
let g i = e:= goto !e i;;
let b () = e:= back !e ;;
let bi i = e:= backi !e i;;
let r () = e:=RdbgStdLib.run();;
let c () = e:= continue !e;;
")^"
open Callgraph;;
let cgf () = gen_call_graph_full !e;;
let cg () = (gen_call_graph !e);;
let dcg ()= ignore(display_call_graph ());;  
let bt () = print_src !e;;
let finish () = loopforever !e;;
"

let _ =
  if Sys.file_exists rdbg_tuning_file then () else
    let oc = open_out rdbg_tuning_file in
    output_string oc (rdbg_tuning());
    flush oc;
    close_out oc
   
                    
let epilog =
  (Printf.sprintf "
let _ =
 args.display_gnuplot <- %s;
 args.display_sim2chro <- %s;
 args.rdbg <- true;
 args.step_nb <- %i;
 args.output <- \"%s\";
 args.cov_file <- \"lurette.cov\";
 args.stop_on_oracle_error <- true;
 args.debug_rdbg <- %s;;
"
     (if !display_gnuplot then "true" else "false")
     (if !display_sim2chro then "true" else "false")
     !test_length !output_file (if !drdbg then "true" else "false")
  )^
  if !lurette then
    Printf.sprintf
      "args.rdbg <- false;;
let  _ = 
    del_hook  \"print_event\";
    let e = run() in
       loopforever e;;\n"
  else ("
       
#use \""^rdbg_tuning_file^"\";;

let _ = print_string \"\n --> type 'man' for online help\n\" 
")

(* Fill the session file and returns the list of used plugins  *)
let gen_rdbg_session_content oc =
  let give_unique_name (i, acc_names, acc_calls) str =  
      let name = Printf.sprintf "plugin_%i" i in
      (i+1, name::acc_names, str::acc_calls)
  in
  let i, sut_names, (sut_calls:string list) = 
    List.fold_left give_unique_name (0, [],[]) !suts
  in
  let i, sut_names_nd, sut_calls_nd = 
    List.fold_left give_unique_name (i, [],[]) !suts_nd
  in
  let i, env_names, env_calls = 
    List.fold_left give_unique_name (i,[],[]) !envs
  in
  let i, env_names_nd, env_calls_nd = 
    List.fold_left give_unique_name (i,[],[]) !envs_nd
  in
  let i, oracle_names, oracle_calls = 
    List.fold_left give_unique_name (i,[],[]) !oracles
  in
  let _, oracle_names_nd, oracle_calls_nd = 
    List.fold_left give_unique_name (i,[],[]) !oracles_nd
  in
  let sut_names = List.rev sut_names in
  let sut_names_nd = List.rev sut_names_nd in
  let env_names = List.rev env_names in
  let env_names_nd = List.rev env_names_nd in
  let oracle_names = List.rev oracle_names in
  let oracle_names_nd = List.rev oracle_names_nd in
  let plugRef = ref [] in
  output_string oc prolog;
  List.iter2 (gen_plugin_def plugRef oc false) sut_names sut_calls;
  List.iter2 (gen_plugin_def plugRef oc true ) sut_names_nd sut_calls_nd;
  List.iter2 (gen_plugin_def plugRef oc false) env_names env_calls;
  List.iter2 (gen_plugin_def plugRef oc true ) env_names_nd env_calls_nd;
  List.iter2 (gen_plugin_def plugRef oc false) oracle_names oracle_calls;
  List.iter2 (gen_plugin_def plugRef oc true ) oracle_names_nd oracle_calls_nd;
  gen_arg_fill oc "suts" (sut_names@sut_names_nd);
  gen_arg_fill oc "envs" (env_names@env_names_nd);
  gen_arg_fill oc "oracles" (oracle_names@oracle_names_nd);
  output_string oc epilog;
  flush oc;
  !plugRef


(**********************************************************************************)

let rdbg_cmd = ref ""
                                             
let rec my_read_line () =
  let str = read_line () in
  match str with
    | "#auto"  -> 
      add_semi_ref:= true;
      print_string "type '#auto_off' to disable automatic semicolumn mode\n";
      flush stdout;
      my_read_line ()
    | "#auto_off" -> 
      add_semi_ref:= false; 
      print_string "type '#auto' to enable automatic semicolumn mode\n";
      flush stdout;
      my_read_line ()
    | _ -> str

let rec myloop oc str =
  let str = if str = "" then my_read_line () else str in
  let str = 
    (* a clutch I'd prefer to avoid *)
    if str = "man" then "RdbgMain.man" else str in
  let add_par = if String.contains str ' ' then "" else "()" in
  let no_semicol = not (Str.string_match (Str.regexp ";;[ \t]+$") str 0) in
  let semi = if !add_semi_ref && no_semicol && 
                  str <> "" && str.[0] <> ' ' then ";;" else "" in
  if str = "q" || str = "quit" then (
    ignore (Unix.close_process_out oc); exit 0) 
  else (
    rdbg_cmd := str;
    let str = str ^ add_par ^ semi in
    if !verbose then (
      output_string stdout ("rdbg sends to the ocaml toplevel the string: \"" ^
                              str ^ "\"\n");
      flush stdout
    );
    output_string oc (str ^ "\n");
    flush oc;
    let str = my_read_line () in
    let str = if str = "" then !rdbg_cmd else str in
    myloop oc str
  ) 


let ls str = Mypervasives.run ("ls -t "^str) (fun s -> Some s)

let _ =   
  try
    if !verbose then (
      print_string ("  "^Sys.argv.(0)^" communicates with the process \""^
                      rdbg_top_cmd^"\"\n"); 
      flush stdout;
      flush stdout
    );
    let call_myloop str =
      let oc = Unix.open_process_out rdbg_top_cmd in
      myloop oc str
    in
    let session_file_name = if !lurette then "luretteSession" else "rdbg-session" in
    let create_new_session () =
      let i = ref 0 in
      let str = ref "" in
      while (Sys.file_exists (session_file_name^ !str^".ml")) do 
        incr i;
        str := (string_of_int !i);
      done;
      let session_file_name = session_file_name^ !str in
      let rdbg_session = session_file_name^".ml" in
      let foc = open_out rdbg_session in
      let plugins = gen_rdbg_session_content foc in
      close_out foc;
      print_string ("  - The file \""^rdbg_session ^ "\" has just been generated. \n");
      flush stdout;
      session_file_name,rdbg_session,plugins
    in
    if other_args = "" then (
      if !lurette then (
        let session_file_name,_ns,plugins = create_new_session () in
        let plugins = "rdbg-plugin"::plugins in
        let plugins = String.concat "," plugins in
        let ocamlopt = try Unix.getenv "OCAMLOPT" with Not_found -> "ocamlopt" in
        let cmd =
          (*   
           The following is a hacky try to see if it is possible to deliver a binary 
           distribution of rbdg/lurette. But it appears that the same exact version
           of  ocamlopt is necessary anyway. sigh.
           *)
         try
          let libdir = Unix.getenv "V6TOOLS_LIBPATH" in
            let cmxa = Unix.getenv "V6TOOLS_CMXA" in
            Printf.sprintf
              "%s -shared -o %s.cmxs %s %s %s.ml\n" ocamlopt
              session_file_name libdir cmxa  session_file_name
        with Not_found ->
            Printf.sprintf
              "ocamlfind opt -shared -o %s.cmxs -package %s -linkpkg %s.ml\n"
              session_file_name plugins session_file_name
        in
        let cmd2 = Printf.sprintf "rdbg-batch %s.cmxs\n" session_file_name in
        let msg = Printf.sprintf " 
 In order to re-run the test using do:
      \027[00m%s\027[32m
 If you tune %s.ml, you need to update %s.cmxs:
      \027[00m%s"
                                 cmd2 session_file_name session_file_name cmd
        in        
        Printf.printf "\027[32m %s \027[00m\n" msg ;
        let err_code = Sys.command cmd in
        let err_code = if err_code > 0 then err_code else Sys.command cmd2 in
        flush stdout;
        ignore (exit err_code)
      ) else (
        let ml_files = ls "rdbg*.ml" in
        let ml_files = List.sort compare ml_files in
        let dir = List.mapi (fun i f -> Printf.sprintf "\n[%i] #use \"%s\"" i f)
                            ml_files
        in
        let dir = dir@["\n[c] create a fresh session\n";
                       "[q] quit\n";"[s] do nothing\n"] in
        let dir = (String.concat "" dir)^ "\n" in
        let info() = print_string dir in
        info();
        print_string ("[0/s/q/c/1/2/...]: ");
        let c = read_line() in
        if List.mem c ["c"] || ml_files = [] && c = "" then (
          let _,ns,_ = create_new_session () in
          let ns = Printf.sprintf "#use \"%s\"\n" ns in
          print_string ns;
          flush stdout;
          call_myloop ns
        ) else
          if List.mem c ["s"] then (
            call_myloop ""
          ) else
            if List.mem c ["";"l";"L"] then (
              let last_ml = List.hd ml_files in
              let last = Printf.sprintf "#use \"%s\"\n" last_ml in
              print_string (last);
              flush stdout;
              call_myloop last
            ) else
              if List.mem c ["q"] then exit 0
              else
                try
                  let i = int_of_string c in
                  let f = List.nth ml_files i in
                  let str = Printf.sprintf "#use \"%s\"\n" f  in
                  print_string str;
                  flush stdout;
                  call_myloop str
                with _ -> 
                  call_myloop  ""
      );
      call_myloop ""
    )
  with 
  | End_of_file -> Printf.printf "bye.\n"; flush stdout
  | e  -> 
     Printf.eprintf "Error in rdbg: %s\n" (Printexc.to_string e);
     flush stderr
